signature PROP =
sig

  structure Ctx : CTX
  structure Term : TERM
    sharing Term.Ctx = Ctx

					(* Propositions   *)
  datatype prop =			(* A ::=          *)
    Atom of string * Term.term list	(*   P(t1,...,tn) *)
  | Tensor of prop * prop		(* | A1 * A2      *)
  | One					(* | 1            *)
  | Lolli of prop * prop		(* | A1 -o A2     *)
  | With of prop * prop			(* | A1 & A2      *)
  | Top					(* | T            *)
  | Plus of prop * prop			(* | A1 + A2      *)
  | Zero				(* | 0            *)
  | Implies of prop * prop		(* | A1 => A2     *)
  | Bang of prop			(* | ! A          *)
  | Forall of Term.var * prop		(* | forall x. A  *)
  | Exists of Term.var * prop		(* | exists x. A  *)

  (* subst (t,x) A ===> [t/x] A; t closed *)
  val subst : Term.term * Term.var -> prop -> prop

  (* eq (A, B) ===> true  iff  A = B (modulo variable names); A and B closed *)
  val eq : prop * prop -> bool

  (* closed P A ===> true  iff  A closed wrt P *)
  val closed : Term.parm Ctx.ctx -> prop -> bool

  val toString : prop -> string

end;  (* signature PROP *)
