signature RULES =
sig

  structure Seq : SEQ

  type hyp = int			(* Hypothesis H *)
  type hyps = hyp list			(* Hypotheses Hs *)
  type split = hyps * hyps		(* Split (Hs1, Hs2) *)

  (*
     A rule r : rule = seq -> goal maps a sequent to subgoals g : goal,
     one for each premiss.  This could be Success (no subgoals), Seq
     (...) (single subgoal), or Conj (g1, g2) (two subgoals).

     Hypotheses H : hyp are numbered in the reverse order they were
     made: most recent is 1, etc.  Thus labels are not kept explicitly
     in the linear and unrestricted contexts, but hypotheses are
     identified by counting.  Unrestricted and linear hypotheses are
     numbered separately.  Note that new hypotheses are always added as
     number 1 (or numbers 2 and 1 in some rules).

     A split (Hs1, Hs2) : split must be such that each hypothesis of the
     conclusion of appears in either Hs1 or Hs2, but not both.

     Example:

      P; G; (D1, D1') ==> A   P; G; (D2, D2', B) ==> C
     -------------------------------------------------- -oL
         P; G; (D1 x D2, A -o B, D1' x D2') ==> C

     is called with

       lolliL (Hs1, Hs2) H Seq

     where Hs1 collects all labels in D1 and D1' in any order,
           Hs2 collects all labels in D2 and D2' in any order,
           H is the label of the hypothesis A -o B
     and 1 will be the implicit label of B in the second subgoal.

     Hypothesis in (D1, D1') and (D2, D2') occur in the same order in
     which they occurred in the conclusion sequent.

     All left rules identify the hypothesis they apply to.
  *)

  val initial : Seq.rule
  val derelict : hyp -> Seq.rule

  val tensorR : split -> Seq.rule
  val tensorL : hyp -> Seq.rule

  val lolliR : Seq.rule
  val lolliL : split -> hyp -> Seq.rule

  val oneR : Seq.rule
  val oneL : hyp -> Seq.rule

  val withR : Seq.rule
  val withL1 : hyp -> Seq.rule
  val withL2 : hyp -> Seq.rule

  val topR : Seq.rule
  (* no topL rule *)

  val plusR1 : Seq.rule
  val plusR2 : Seq.rule
  val plusL : hyp -> Seq.rule

  (* no zeroR rule *)
  val zeroL : hyp -> Seq.rule

  val impliesR : Seq.rule
  val impliesL : hyp -> Seq.rule

  val bangR : Seq.rule
  val bangL : hyp -> Seq.rule

  (* forallR a : rule; a new wrt parameter context P *)
  val forallR : Seq.Prop.Term.parm -> Seq.rule

  (* forallL t H : rule; t closed wrt parameter context P *)
  val forallL : Seq.Prop.Term.term -> hyp -> Seq.rule

  (* existsR t : rule; t closed wrt parameter context P *)
  val existsR : Seq.Prop.Term.term -> Seq.rule

  (* existsL a H : rule; a new wrt parameter context P *)
  val existsL : Seq.Prop.Term.parm -> hyp -> Seq.rule

end;  (* signature RULES *)
