functor Seq
  (structure Ctx' : CTX
   structure Prop' : PROP
     sharing Prop'.Ctx = Ctx')
  : SEQ
=
struct

  structure Ctx = Ctx'
  structure Prop = Prop'

  local
    open Ctx
    infix $
  in

    (* Sequents [P] G; D ==> C *)
    (* Invariant:
       All parameters in G, D, and C are declared in P
       Propositions in G, D, and C contain no free variables
    *)
    type seq = Prop.Term.parm Ctx.ctx	(* [P]   *)
	       * Prop.prop Ctx.ctx	(* G;    *)
	       * Prop.prop Ctx.ctx	(* D     *)
	       * Prop.prop		(* ==> C *)

    fun parmsToString P = "[" ^ pts P ^ "]"
    and pts (Null) = ""
      | pts (Null $ p) = p
      | pts (ps $ p) = pts ps ^ "," ^ p

    fun ctxToString (Null, i) = "Null"
      | ctxToString (Null $ A, i) = Int.toString i ^ " : " ^ Prop.toString A
      | ctxToString (G $ A, i) =
	  ctxToString (G, i+1) ^ ",\n"
	  ^ Int.toString i ^ " : " ^ Prop.toString A

    fun toString (P, G, D, C) =
	parmsToString P ^ "\n"
	^ ctxToString (G, 1) ^ ";\n"
	^ ctxToString (D, 1) ^ "\n"
	^ "==>  " ^ Prop.toString C ^ "\n"
	^ "--------------------------------------------------"

    (* Goals *)
    (* Invariant:
       Success never appears as a conjunct 
    *)
    datatype goal =
      Success
    | Conj of goal * goal
    | Seq of seq

    (* exception Fail(reason) provides a brief message *)
    exception Fail of string

    (* A rules reduces a goal sequent to subgoals *)
    type rule = seq -> goal

  end  (* local ... *)
end;  (* functor Seq *)
