signature SEQ =
sig

  structure Ctx : CTX
  structure Prop : PROP
    sharing Prop.Ctx = Ctx

  (*
     A sequent P; G; D ==> C consists of
     a parameter context P which contains all individual parameters
       which may occur in G, D, and C,
     a context G of unrestricted hypotheses,
     a context D of linear hypothesis,
     and a proposition C.
  *)
  type seq = Prop.Term.parm Ctx.ctx	(* [P]   *)
             * Prop.prop Ctx.ctx	(* G;    *)
             * Prop.prop Ctx.ctx	(* D     *)
             * Prop.prop		(* ==> C *)

  val toString : seq -> string

  (*
     A goal is a conjunction of the sequents which remain
     to be derived.  Success represents the empty conjunction
     (that is, no goals remain), and Seq represent a single
     goal.

     We maintain the invariant that Success does not appear as a
     conjunct in a goal structure.
  *)
  datatype goal =
    Success
  | Conj of goal * goal
  | Seq of seq

  (*
    exception Fail(reason) is raised when an inference rule can be
    applied in the backward direction to reduce the current goal to
    subgoals.
  *)
  exception Fail of string

  (* A rule reduces a sequent to subgoals *)
  (* Currently we do not attempt to validate inferences! *)
  type rule = seq -> goal

end;  (* signature SEQ *)
