functor Top
  (structure Seq' : SEQ)
  : TOP =
struct

  structure Seq = Seq'

  local
    open Seq.Ctx
    infix $
    structure Prop = Seq.Prop

    val goalref = ref Seq.Success
    val judgref = ref (NONE:Seq.seq option)
    exception Error

    exception Solved
    fun first (Seq.Success) = raise Solved
      | first (Seq.Conj(g1,g2)) = first g1
      | first (Seq.Seq seq) = seq

    fun showFirstGoal (g) =
        print (Seq.toString (first g) ^ "\n")
	handle Solved => print "*** Goal solved ***\n"

    fun showGoals (Seq.Success) = print "*** Subgoal solved ***\n"
      | showGoals (Seq.Conj (g1, g2)) =
        (showGoals g2;
	 showGoals g1)
      | showGoals (Seq.Seq (seq)) = print (Seq.toString seq ^ "\n")

    fun conjoin (Seq.Success, g2) = g2
      | conjoin (g1, Seq.Success) = g1
      | conjoin (g1, g2) = Seq.Conj (g1, g2)

    fun applyRule (r, Seq.Success) =
        (print ("No subgoal left\n"); raise Error)
      | applyRule (r, Seq.Conj(g1,g2)) = conjoin (applyRule (r, g1), g2)
      | applyRule (r, g as Seq.Seq seq) =
        let val g' = r seq
	    val _ = showGoals g'
	in g' end
        handle Seq.Fail (msg) => (print ("Fails: " ^ msg ^ "\n"); raise Error)
  in
    exception Error = Error
    fun show () = showFirstGoal (!goalref)
    fun showAll () = showGoals (!goalref)

    fun prove (A) =
        let
	  val _ = if Prop.closed Null A
		    then ()
		  else (print ("Proposition not closed\n"); raise Error)
	  val seq = (Null, Null, Null, A); 
	  val _ = (goalref := Seq.Seq seq)
	  val _ = (judgref := SOME seq)
	in
	  show ()
	end
        (* handle Error => () *)

    fun entail (A,B) =
        let
	  val _ = if Prop.closed Null A andalso Prop.closed Null B
		    then ()
		  else (print ("Sequent not closed\n"); raise Error)
	  val seq = (Null, Null, Null $ A, B)
	  val _ = (goalref := Seq.Seq seq)
	  val _ = (judgref := SOME seq)
	in
	  show ()
	end
        (* handle Error => () *)

    fun apply (rule) =
        (goalref := applyRule (rule, !goalref))
	(* handle Error => () *)

    fun judgment () =
        case (!goalref, !judgref)
	  of (Seq.Success, SOME(seq)) =>
	        (print ("*** Judgment ***\n");
		 print (Seq.toString seq ^ "\n");
		 print ("*** evident ***\n");
		 true)
           | (_, SOME(seq)) =>
	        (print ("*** Judgment ***\n");
		 print (Seq.toString seq ^ "\n");
		 print ("*** not yet evident ***\n");
		 false)
           | (_, NONE) =>
		(print ("*** No derivation in progress ***\n");
		 false)

    fun done () =
        case (!goalref)
	  of Seq.Success => ()
	   | _ => raise Error

  end  (* local ... *)
end;  (* functor Top *)
