signature TOP =
sig

  structure Seq : SEQ

  (*
     The top level is state-oriented and maintains a current
     goal structure, the left-most subgoal of which is the current
     goal sequent.  A derivation is initialized with the
     prove and entail commands.
  *)

  (* Propositions A and B must be closed and may not contain parameters *)
  val prove : Seq.Prop.prop -> unit	(* [.] .; . |- A ? *)
  val entail : Seq.Prop.prop * Seq.Prop.prop -> unit
					(* [.] .; A |- B ? *)

  exception Error

  val show : unit -> unit		(* current goal sequent *)
  val showAll : unit -> unit		(* all current subgoals *)

  val apply : Seq.rule -> unit		(* apply rule to current goal *)

  val judgment : unit -> bool		(* print evident judgment (if any) *)
  val done : unit -> unit		(* check if judgment is evident *)

end; (* signature TOP *)
